#%%
import obspy as op
from obspy.core import read
from obspy.taup import TauPyModel
from obspy.geodetics import locations2degrees
from obspy.clients.fdsn import Client

from pyrocko import cake

import numpy as np
import scipy as sp
import matplotlib.pyplot as plt

import multiprocessing
from functools import partial

import sys
import os
import re
import json
from pathlib import Path
import logging
import argparse    

log = logging.getLogger(__name__)

#%%
def load_jsonc(filepath):
    with open(filepath, 'r') as f:
        text = f.read()
    text_without_comment = re.sub(r'/\*[\s\S]*?\*/|//.*', '', text)
    json_obj = json.loads(text_without_comment)
    return json_obj

def aic(td):
    """
    Differently from AR-AIC picker, this function calculates AIC function
    directly from the data, without using the AR coefficients (Maeda, 1985)
    Computes P-phase arrival time digital single-component acceleration
    or broadband velocity record without requiring threshold settings using
    AKAIKE INFORMATION CRITERION.
    Returns P-phase arrival time index and the characteristic function.
    The returned index correspond to the characteristic function's minima.
    :type td: :class:`numpy.ndarray`
    :param td: time series values.
    :return: (idx, aic_cf) idx sample number of parrival; aic_cf numpy.ndarray
        containing the values of the characteristic function.
    .. seealso:: [Maeda1985]_
    """
    # --------------------  Initial checks
    if not isinstance(td, np.ndarray):
        raise ValueError("Input time series must be a numpy.ndarray instance!")
    aic_cf = np.zeros(td.size - 1)
    # --------------------  CF calculation
    with np.errstate(divide='raise'):
        for ii in range(1, td.size):
            try:
                var1 = np.log(np.var(td[0:ii]))
            except FloatingPointError:  # if var==0 --> log is -inf
                var1 = 0.00
            #
            try:
                var2 = np.log(np.var(td[ii:]))
            except FloatingPointError:  # if var==0 --> log is -inf
                var2 = 0.00
            #
            val1 = ii * var1
            val2 = (td.size - ii - 1) * var2
            aic_cf[ii - 1] = (val1 + val2)
    # --------------------  Index search
    idx = np.argmin(aic_cf)
    return idx, aic_cf

def aic_polt(tr, idx, aic_cf):
    plt.rcParams['figure.figsize'] = 15, 7
    ax = plt.subplot(211)
    plt.plot(tr.data, 'k')
    ymin, ymax = ax.get_ylim()
    plt.vlines(idx, ymin, ymax, color='r', linewidth=2)
    ax2 = plt.subplot(212, sharex=ax)
    plt.plot(aic_cf, 'k')
    ymin, ymax = ax2.get_ylim()
    print(ymin)
    plt.vlines(idx, ymin, ymax, color='r', linewidth=2)
    plt.show()
    
def sac_header_info(tr):
    evla = tr.stats.sac['evla']
    evlo = tr.stats.sac['evlo']
    evdp = tr.stats.sac['evdp']

    evid = tr.stats.sac['kevnm']
    mag = tr.stats.sac['mag']

    try:
        ev_origin = tr.stats.sac['o']
    except:
        ev_origin = None

    time_b = tr.stats.sac['b']
    time_e = tr.stats.sac['e']

    stla = tr.stats.sac['stla']
    stlo = tr.stats.sac['stlo']

    
    return evla, evlo, evdp, mag, evid, stla, stlo, ev_origin, time_b, time_e

def cal_parrival(model, distance_km, source_depth_km):
    kmtom = 1000
    #distance_km = np.array([0])
    #distance_km = np.linspace(0, 10, 11)*km * cake.m2d

    source_depth_m = source_depth_km * kmtom
    log.info("source_depth_km = {0}".format(source_depth_km))
    log.info("source_depth_m = {0}".format(source_depth_m))   
    
    distance_deg = np.array([distance_km]) * kmtom * cake.m2d
    log.info("distance_km = {0}".format(distance_km))
    log.info("distance_deg = {0}".format(distance_deg))

    Phase = "p"

    try:
        arrival = model.arrivals(distance_deg, phases=Phase, zstart=source_depth_m)
        if len(arrival) == 0:
            log.debug("no p phase for distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
            Phase = "P"
            arrival = model.arrivals(distance_deg, phases=Phase, zstart=source_depth_m)
            if len(arrival) == 0:
                log.debug("no both p and P phase for distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
    except:
        log.warning("no p & P phase computefor distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
        arrival = None


    return arrival

#%%
def do_preprocess(vel_q_nd_fi, eq_client_name, ev_info_againOPT, sampling_rate_ref,
                  pre_tw_aic, post_tw_aic, fl_aic, fh_aic, pre_tw_save, post_tw_save, fl_xc, fh_xc, 
                  output_top_dir, plotOPT,
                  task):

    st1_fi = task
    #sacfi_name = st1_fi.name.split('/')[-1]
    sacfi_name = st1_fi.split('/')[-1]


    try:
        st1 = read(st1_fi) 
        tr1 = st1[0]
        tr1_idOUT = tr1.id.replace('.', '_')
        #cs_xc.BKS.BK.HHZ.00.all
        output_dir = output_top_dir+"/cs_xc."+tr1.id+".all/"+tr1_idOUT+"_"+str(fl_xc)+"_"+str(fh_xc)+"_"+str(pre_tw_save)+"_"+str(post_tw_save)+"_data.all"
        log.info("output_dir = {0}".format(output_dir))

        p_output_dir = Path(output_dir)

        ### outout_dir was created before do_processing. some errors during parallel computing. simultabeously made directory...
        #if p_output_dir.exists() is not True:
        #    log.debug("will make p_output_dir = {0}".format(p_output_dir))
        #    p_output_dir.mkdir(parents=True)
        ### 
        
        #log.debug(output_dir)
        #log.debug(type(output_dir))
        #log.debug("type(st1_fi) = {0}".format(type(st1_fi)))
        output_fi = output_dir+"/"+sacfi_name
        #log.debug("st1_fi = {0}".format(st1_fi))
        log.debug("output_fi = {0}".format(output_fi))


        log.debug(tr1.stats)
        
        if plotOPT:
            tr1.plot()
        
        evla1, evlo1, evdp1, mag1, evid1, stla1, stlo1, ev_origin1, time_b1, time_e1 = sac_header_info(tr1)
        stdp1 = 0.0 # km 
        log.debug("evla1 = {0} evlo1 = {1} evdp1 = {2}".format(evla1, evlo1, evdp1))
        log.debug("mag1 = {0} evid1 = {1}".format(mag1, evid1))
        log.debug("stla1 = {0} stlo1 = {1}".format(stla1, stlo1))
        log.debug("ev_origin1 = {0} time_b1 = {1} time_e1 = {2}".format(ev_origin1, time_b1, time_e1))


        if ev_origin1 is None or ev_info_againOPT:
            eq_found = 0 

            log.warning("check ev_info again evid1 {0}".format(evid1))
            eq_client = Client(eq_client_name)
            try:
                log.debug("hypoDD catalog evid1 = {0}".format(evid1))
                # this only works for NCEDC catalog="DD". we can use eq_client_list dictionary but too much now. all data should have eq info.
                catalog = eq_client.get_events(eventid=evid1, catalog="DD")
                eq_found = 2 

            except:
                log.warning("NCSS catalog evid1 = {0}".format(evid1))
                try:
                    catalog = eq_client.get_events(eventid=evid1)
                    eq_found = 1 

                except:     
                    log.warning("no catalog evid1 = {0}".format(evid1))

            if eq_found >= 1:
                log.debug("new event info evid1 = {0}".format(evid1))
                
                event = catalog[0]
                origin = event.origins[0]
                origin_time = origin.time
                evla1 = origin.latitude
                evlo1 = origin.longitude
                evdp1 = origin.depth / 1000
                try:
                    mag1 = event.magnitudes[0].mag 
                except:
                    log.warning("evid1 {0} has no mag info we use mag=-9".format(evid1))
                    mag1 = -9 # there is a case that we have mag before but now (2023), NCSS will not have mag info.
                ev_origin1 = origin_time - tr1.stats.starttime
                log.debug("evla1 = {0} evlo1 = {1} evdp1 = {2}".format(evla1, evlo1, evdp1))
                log.debug("mag1 = {0} evid1 = {1}".format(mag1, evid1))
                log.debug("ev_origin1 = {0} ".format(ev_origin1))
                tr1.stats.sac['evla'] = evla1
                tr1.stats.sac['evlo'] = evlo1
                tr1.stats.sac['evdp'] = evdp1
                tr1.stats.sac['mag'] = mag1
                tr1.stats.sac['o'] = ev_origin1

                if eq_found == 1:
                    #tr1.stats.sac['kuser0'] = "ncss"
                    tr1.stats.sac['kuser0'] = "default"

                if eq_found == 2:
                    #tr1.stats.sac['kuser0'] = "hypoDD"
                    tr1.stats.sac['kuser0'] = "relocation"
                
                log.debug("sac header again = {0}".format(tr1.stats.sac))

            else:
                log.debug("cano")
                log.error("cannot find evid1 {0} from eq_client_name {1} eq_found {2}".format(evid1, eq_client_name, eq_found))





        distance_deg = locations2degrees(evla1, evlo1, stla1, stlo1)
        distance_km = (cake.d2m/1000.0)*distance_deg

        source_depth_km = evdp1
        if source_depth_km < 0:
            # evdp1 is geoide depth. currently igorno datum. avbout 600m for HF region. so some EQ would be negative and we forced depth to be zero for P arrival time
            log.warning("evid1 {0} is negative depth {1}. we use depth=0 for P arraival".format(evid1, evdp1))
            source_depth_km = 0

        #distance_km = 1
        model = cake.load_model(fn=vel_q_nd_fi, format='nd')
        arrival = cal_parrival(model, distance_km, source_depth_km)

        arrival_found = 0
        if arrival is not None:
            log.info("arrival = {0}".format(arrival))
            log.info("arrival[0].t = {0}".format(arrival[0].t))
            tr1.stats.sac['a'] = tr1.stats.sac['o'] + arrival[0].t
            arrival_found = 1
        else:
            log.warning("no arrival found tr1.id = {0} evid1 = {1}".format(tr1.id, evid1))

        log.debug(tr1)
        tr1_resamp = tr1.copy().detrend("demean").detrend("linear").resample(sampling_rate_ref)
        log.debug(tr1_resamp)
        
        taper = 0.005
        corners = 6

        if arrival_found:
            log.debug("move aic part")
            tr1_resamp_fil = tr1_resamp.copy().detrend("demean").detrend("linear").taper(taper).filter("bandpass", 
                        freqmin=fl_aic, freqmax=fh_aic, zerophase=False, corners=corners)
            
            if plotOPT:
                tr1_resamp_fil.plot()

            cut_st = tr1.stats.starttime + tr1_resamp_fil.stats.sac['a'] - pre_tw_aic
            cut_et = tr1.stats.starttime + tr1_resamp_fil.stats.sac['a'] + post_tw_aic
            log.debug("cut_st = {0}".format(cut_st))
            log.debug("cut_et = {0}".format(cut_et))

            aic_cut_found = 0
            try:
                tr1_resamp_fil_cut = tr1_resamp_fil.copy().trim(starttime=cut_st, endtime=cut_et)
                aic_cut_found = 1
            except Exception as e:
                log.warning("no enough data for aic pick {0} npts = {1}".format(sacfi_name, tr1_resamp_fil_cut.stats.npts))
                log.warning(e)


            log.debug("tr1_resamp_fil_cut = {0}".format(tr1_resamp_fil_cut))

            if tr1_resamp_fil_cut.stats.npts == 0:
                 log.warning("no enough data for aic pick {0} npts = {1}".format(sacfi_name, tr1_resamp_fil_cut.stats.npts))
                 aic_cut_found = 0


            if aic_cut_found:
                if plotOPT:
                    tr1_resamp_fil_cut.plot()
                aic_found = 0
                idx, aic_cf = aic(tr1_resamp_fil_cut.data)
                if plotOPT:
                    aic_polt(tr1_resamp_fil_cut, idx, aic_cf)

                if idx >= 0:
                    aic_found = 1
                else:
                    log.warning("no aic found sacfi = {0}".format(st1_fi))

                    t5_omitted_dir = output_fi = output_dir+"/t5omitted"
                    p_t5_omitted_dir = Path(t5_omitted_dir)

                    ### directory was created before do_preprocessing
                    #if p_t5_omitted_dir.exists() is not True:
                    #    log.debug("will make p_t5_omitted_dir = {0}".format(p_t5_omitted_dir))
                    #    p_t5_omitted_dir.mkdir()
                    ###

                    tr1_resamp.write(t5_omitted_dir+"/"+sacfi_name, format="SAC")


                if aic_found:
                    aic_tt = tr1_resamp_fil_cut.stats.starttime  + tr1_resamp_fil_cut.stats.delta * idx
                    log.debug("aic_tt = {0}".format(aic_tt))    
                    aic_time_sac_header = aic_tt - tr1_resamp.stats.starttime 
                    log.debug("aic_time_sac_header = {0}".format(aic_time_sac_header))    
                    tr1_resamp.stats.sac['t5'] = aic_time_sac_header
                    
                    cut_save_st = tr1_resamp.stats.starttime + tr1_resamp.stats.sac['t5'] - pre_tw_save
                    cut_save_et = tr1_resamp.stats.starttime + tr1_resamp.stats.sac['t5'] + post_tw_save
                
                    tr1_resamp_fil_save = tr1_resamp.copy().detrend("demean").detrend("linear").taper(taper).filter("bandpass", 
                        freqmin=fl_xc, freqmax=fh_xc, zerophase=False, corners=corners).trim(starttime=cut_save_st, endtime=cut_save_et)

                    #tr1_resamp_fil_save.write("tr1_resamp_fil_save.sac", format="SAC")
                    #output_fi = output_dir+"/"+st1_fi
                    #log.debug("output_fi = {0}".format(output_dir+"/"+st1_fi))
                    tr1_resamp_fil_save.write(output_fi, format="SAC")



    except Exception as e: 
        log.error("cannot preprocess for {0}".format(sacfi_name))
        log.error(e)


#%%
def start_parallel_jobs_inner_loop(tasks, do_work, njobs=0):
    if njobs == 0:
        results = [do_work(task) for task in tasks]
    else:
        pool = multiprocessing.Pool(njobs)
        results = pool.map(do_work, tasks)
        pool.close()
        pool.join()
        
    return results
    
#%%
def preprocess(data_top_dir, trace_id, traces_dict,
               vel_q_nd_fi, eq_client_name, ev_info_againOPT, sampling_rate_ref,
               pre_tw_aic, post_tw_aic, fl_aic, fh_aic, pre_tw_save, post_tw_save, fl_xc, fh_xc, output_top_dir,
               plotOPT, njobs, task_limit):



    log.info("data_top_dir = {0}".format(data_top_dir))
    log.info("vel_q_nd_fi = {0}".format(vel_q_nd_fi))

    log.info("eq_client_name = {0}".format(eq_client_name))
    log.info("ev_info_againOPT = {0}".format(ev_info_againOPT))
    log.info("sampling_rate_ref = {0}".format(sampling_rate_ref))

    log.info("pre_tw_aic = {0}".format(pre_tw_aic))
    log.info("post_tw_aic = {0}".format(post_tw_aic))

    log.info("fl_aic = {0}".format(fl_aic))
    log.info("fh_aic = {0}".format(fh_aic))
    
    log.info("pre_tw_save = {0}".format(pre_tw_save))
    log.info("post_tw_save = {0}".format(post_tw_save))

    log.info("fl_xc = {0}".format(fl_xc))
    log.info("fh_xc = {0}".format(fh_xc))

    log.info("output_top_dir = {0}".format(output_top_dir))
    log.info("plotOPT = {0}".format(plotOPT))

    log.info("njobs = {0}".format(njobs))
    log.info("task_limit = {0}".format(task_limit))
    
    netOUT, staOUT, locOUT, comOUT = trace_id.split('.')
    log.debug("netOUT = {0} staOUT = {1} locOUT = {2} comOUT = {3}".format(staOUT, netOUT, locOUT, comOUT))
    trace_idOUT = trace_id.replace('.', '_')
    #cs_xc.BKS.BK.HHZ.00.all
    output_dir = output_top_dir+"/cs_xc."+trace_id+".all/"+trace_idOUT+"_"+str(fl_xc)+"_"+str(fh_xc)+"_"+str(pre_tw_save)+"_"+str(post_tw_save)+"_data.all"
    log.info("output_dir = {0}".format(output_dir))

    p_output_dir = Path(output_dir)

    ### outout_dir was created before do_processing. some errors during parallel computing. simultabeously made directory...
    if p_output_dir.exists() is not True:
        log.debug("will make p_output_dir = {0}".format(p_output_dir))
        p_output_dir.mkdir(parents=True)

    ### t5 omitted
    t5_omitted_dir = output_fi = output_dir+"/t5omitted"
    p_t5_omitted_dir = Path(t5_omitted_dir)
    if p_t5_omitted_dir.exists() is not True:
        log.debug("will make p_t5_omitted_dir = {0}".format(p_t5_omitted_dir))
        p_t5_omitted_dir.mkdir()            
    ### t5 omitted

    #sys.exit()

    p = Path(data_top_dir)
    sac_fis = list(p.glob('./*deg/'+staOUT+'.'+netOUT+'.'+comOUT+'.'+locOUT+'*.sac'))
    #('[a-z][a-z][a-z].*'
    log.info("len(sac_fis) = {0}".format(len(sac_fis)))

    # additional traces
    #traces_dict[trace_id]
    trace_ids = traces_dict.get(trace_id)
    log.info("trace_ids = {0}".format(trace_ids))
    log.info("type(trace_ids) = {0}".format(type(trace_ids)))

    if trace_ids is not None:
        for i, trace_id_additional in trace_ids.items():
            log.info("i, trace_id_additional = {0} {1}".format(i, trace_id_additional))

            netOUT, staOUT, locOUT, comOUT = trace_id_additional.split('.')

            log.debug("add netOUT = {0} staOUT = {1} locOUT = {2} comOUT = {3}".format(staOUT, netOUT, locOUT, comOUT))

            trace_id_additionalOUT = trace_id_additional.replace('.', '_')
            #cs_xc.BKS.BK.HHZ.00.all
            output_dir = output_top_dir+"/cs_xc."+trace_id_additional+".all/"+trace_id_additionalOUT+"_"+str(fl_xc)+"_"+str(fh_xc)+"_"+str(pre_tw_save)+"_"+str(post_tw_save)+"_data.all"
            log.info("add output_dir = {0}".format(output_dir))

            p_output_dir = Path(output_dir)

            ### outout_dir was created before do_processing. some errors during parallel computing. simultabeously made directory...
            if p_output_dir.exists() is not True:
                log.debug("add will make p_output_dir = {0}".format(p_output_dir))
                p_output_dir.mkdir(parents=True)

            
            ### t5 omitted
            t5_omitted_dir = output_fi = output_dir+"/t5omitted"
            p_t5_omitted_dir = Path(t5_omitted_dir)
            if p_t5_omitted_dir.exists() is not True:
                log.debug("add will make p_t5_omitted_dir = {0}".format(p_t5_omitted_dir))
                p_t5_omitted_dir.mkdir()            
            ### t5 omitted


            p = Path(data_top_dir)
            sac_fis_additional = list(p.glob('./*deg/'+staOUT+'.'+netOUT+'.'+comOUT+'.'+locOUT+'.*.sac'))

            log.info("total_sac_fis_additional = {0}".format(len(sac_fis_additional)))
            sac_fis = sac_fis + sac_fis_additional

    #log.info("sac_fis = {0}".format(sac_fis))
    log.info("len(sac_fis) = {0}".format(len(sac_fis)))

    if len(sac_fis) > 0:
        log.debug("sac_fis[0].name = {0}".format(sac_fis[0].name))
        log.debug("sac_fis[0].resolve() = {0}".format(sac_fis[0].resolve()))

    #sys.exit()

    tasks = []
    tasks = list(map(lambda x: str((x.resolve()).parents[0]) +"/"+ (x.resolve()).name  , sac_fis))

    #for i, sac_fi in enumerate(sac_fis):
    #    tasks.append(sac_fis[i].resolve())

    #log.debug("tasks = {0}".format(tasks))
    log.info("len(tasks) = {0}".format(len(tasks)))
    if len(tasks) > 0:
        log.debug("tasks[0] = {0}".format(tasks[0]))

    #sys.exit()

    do_work = partial(do_preprocess, vel_q_nd_fi, eq_client_name, ev_info_againOPT, sampling_rate_ref,
                        pre_tw_aic, post_tw_aic, fl_aic, fh_aic, pre_tw_save, post_tw_save, fl_xc, fh_xc, output_top_dir,
                        plotOPT)
    
    if len(tasks) > 0:
        try:
            for i in range(0, len(tasks), task_limit):
                log.info("slice from i {0} to i+task_limit {1} ".format(i, i+task_limit))
                task_subset = (tasks[i: i+task_limit])

                start_parallel_jobs_inner_loop(task_subset, do_work, njobs)

        except Exception as e:
            log.error("cannot preprocess for all data")
            log.error(e)
    

#%%
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='CRE_preprocessing')   
    parser.add_argument('-t', '--trace_id', help='trace id', required=True)   
    parser.add_argument('-p', '--parameter_file', help='parameter file', default="CRE_preprocessing.jsonc")   
    parser.add_argument('-tl', '--traces_list', help='traces list', default="CRE_traces_list.json")   
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()   
    logging_level = "INFO"
    if args.debug:
        logging_level = "DEBUG"
  
    logging.basicConfig(level=logging_level, format="%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    log.info(log)

    #inp_fi = "CRE_preprocessing.json"
    inp_fi = args.parameter_file
    log.info("inp_fi = {0}".format(inp_fi))

    inp_fd = open(inp_fi, mode='r')
    
    #inp_para = json.load(inp_fd)
    inp_para = load_jsonc(inp_fi)
    inp_fd.close()
    log.info('inp_para = {0}'.format(inp_para))


    #traces_list_fi = "CRE_traces_list.json"
    traces_list_fi = args.traces_list
    log.info("traces_list_fi = {0}".format(traces_list_fi))

    traces_dict = {}
    with open(traces_list_fi, mode="r") as f:
        #traces_dict = json.load(f)
        traces_dict = json.load(f)

    log.info("traces_dict = {0}".format(traces_dict))


    #sys.exit()

    #%%
    preprocess(inp_para['data_top_dir'], args.trace_id, traces_dict, inp_para['vel_q_nd_fi'], 
               inp_para['eq_client_name'], inp_para['ev_info_againOPT'], inp_para['sampling_rate_ref'],
               inp_para['pre_tw_aic'], inp_para['post_tw_aic'], inp_para['fl_aic'], inp_para['fh_aic'], 
               inp_para['pre_tw_save'], inp_para['post_tw_save'], inp_para['fl_xc'], inp_para['fh_xc'],
               inp_para['output_top_dir'], inp_para['plotOPT'], inp_para['njobs'], inp_para['task_limit'])

# %%
