#%%
import obspy as op
from obspy.core import read
from obspy.taup import TauPyModel
from obspy.geodetics import locations2degrees

import pandas as pd
import numpy as np

import json
import logging
from pathlib import Path

import multiprocessing
from functools import partial

from pyrocko import cake


logging.basicConfig(
    level=logging.DEBUG,
    #level=logging.INFO,
    #level=logging.WARN,
    #level=logging.ERROR,
    format="%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")

log = logging.getLogger(__name__)
log.info(log)

# %%
def rho_from_vp_ludwig(vp):
    """
    Brocher (2005) equation (1)
    """
    rho = 1.6612*vp
    rho += -0.4721*(vp**2)
    rho += 0.0671*(vp**3)
    rho += -0.0043*(vp**4)
    rho += 0.000106*(vp**5)
    return rho

def rho_from_vp_gardner(vp):
    rho = 1.74*(vp**0.25)
    return rho

def vp_from_rho(rho):
    """
    Brocher (2005) equation (5)
    """
    vp = 39.128*rho
    vp += -63.064*(rho**2)
    vp += 37.083*(rho**3)
    vp += -9.1819*(rho**4)
    vp += 0.8228*(rho**5)
    return vp

def vs_from_vp(vp):
    """
    Brocher (2005) equation (6)
    """
    vs = 0.7858 
    vs += -1.2344*vp
    vs += 0.7949*(vp**2)
    vs += -0.1238*(vp**3)
    vs += 0.0064*(vp**4)
    return vs

def vp_from_vs(vs):
    """
    Brocher (2005) equation (9)
    """
    vp = 0.9409
    vp += 2.0947*vs
    vp += -0.8206*(vs**2)
    vp += 0.2683*(vs**3)
    vp += -0.0251*(vs**4)
    return vp

def qs_from_vs(vs):
    """
    Brocher (2008) equation (19)
    """
    qs = -16 + 104.13*vs
    qs += -25.225*(vs**2)
    qs += 8.2184*(vs**3)
    return qs
def qp_from_qs(qs):
    return 2*qs

def create_model(vel_nd_fi, vel_q_nd_fi, vel_ref_fi):
    
    vel_data = pd.read_csv(vel_nd_fi, sep=" ",names=["dep", "vp", "vs"],header=None)
    log.debug("vel_data = {0}".format(vel_data))
    #vel_ref_data = pd.read_csv("vel_ref_fi", sep='\t',names=["dep", "vp", "vs", "qp", "qs"],header=None)
    # read only 10 (up to 410km) from mantle (depth is 35km). if original vel has more than 35km, then problem...
    vel_ref_data = pd.read_csv("ak135-f-continental.m.nd.csv", sep=',',names=["layer", "dep", "vp", "vs", "rho", "qp", "qs", "dummy"],header=None,  nrows=10, skiprows=5)    
    log.debug("vel_ref_data = {0}".format(vel_ref_data))

    with open(vel_q_nd_fi, 'w') as f:

        for i in range(len(vel_data['vp'])):
            #print(lat,lon,depth)
            #rho_l, rho_g. use rho_l 
            rho_l = rho_from_vp_ludwig(vel_data['vp'][i])
            qs = qs_from_vs(vel_data['vs'][i])
            qp = qp_from_qs(qs)
            f.write('{:10.3f} {:11.4f} {:11.4f} {:11.4f} {:11.4f} {:11.4f}'.format(vel_data['dep'][i],vel_data['vp'][i],vel_data['vs'][i],rho_l,qp,qs)+"\n")
        
        f.write('{0}\n'.format("mantle"))

        for i in range(len(vel_ref_data['vp'])):
            f.write('{:10.3f} {:11.4f} {:11.4f} {:11.4f} {:11.4f} {:11.4f}'.format(vel_ref_data['dep'][i],vel_ref_data['vp'][i],vel_ref_data['vs'][i],vel_ref_data['rho'][i],vel_ref_data['qp'][i],vel_ref_data['qs'][i])+"\n")


def cal_parrival(model, distance_km, source_depth_km):
    kmtom = 1000
    #distance_km = np.array([0])
    #distance_km = np.linspace(0, 10, 11)*km * cake.m2d

    source_depth_m = source_depth_km * kmtom
    log.info("source_depth_km = {0}".format(source_depth_km))
    log.info("source_depth_m = {0}".format(source_depth_m))   
    
    distance_deg = np.array([distance_km]) * kmtom * cake.m2d
    log.info("distance_km = {0}".format(distance_km))
    log.info("distance_deg = {0}".format(distance_deg))

    Phase = "p"

    try:
        arrival = model.arrivals(distance_deg, phases=Phase, zstart=source_depth_m)
        if len(arrival) == 0:
            log.debug("no p phase for distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
            Phase = "P"
            arrival = model.arrivals(distance_deg, phases=Phase, zstart=source_depth_m)
            if len(arrival) == 0:
                log.debug("no both p and P phase for distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
    except:
        log.warning("no p & P phase computefor distance = {0} source_dep_km = {1}".format(distance_km, source_depth_km))
        arrival = None


    return arrival

#%%
if __name__ == '__main__':
    vel_nd_fi = "vel.1D.HF.dat.hypo2000.nd"
    vel_q_nd_fi = "vel.q.1D.HF.dat.hypo2000.nd"
    vel_ref_fi = "ak135-f-continental.m.nd.csv" # reference model used below mantle 

    create_model(vel_nd_fi, vel_q_nd_fi, vel_ref_fi)


    #%%


    #%%
    source_depth_km = 1
    distance_km = 1
    model = cake.load_model(fn=vel_q_nd_fi, format='nd')
    arrival = cal_parrival(model, distance_km, source_depth_km)
    if arrival is not None:
        log.info("arrival = {0}".format(arrival))
        log.info("arrival[0].t = {0}".format(arrival[0].t))



    # %%
